from decimal import Decimal
from typing import List

from .constants import ELSS, EQUITY
from .models import Scheme
from .transaction_utils import get_purchase_transactions_for_active_units, get_filtered_transactions
from .transaction_filters import get_transaction_older_than_filter
from .scheme_filters import SchemeFilterType
from .utils import logger


def get_filtered_schemes(filter_to_apply: SchemeFilterType, schemes: List[Scheme]):
    """Apply filter on schemes and return the filtered list"""
    return list(filter(filter_to_apply, schemes))


def analyze_ltcg_tax_harvesting(scheme: Scheme):
    """Analyze Long Term Capital Gains opportunities

    Args:
        scheme (Scheme):

    Returns:
        dict: ltgc stats for given scheme
    """
    logger.debug(f"LTCG Tax Harvesting for Scheme: {scheme.name}")

    if scheme.type == EQUITY:
        if scheme.subtype == ELSS:
            num_days = 365 * 3
        else:
            num_days = 365
    else:
        logger.debug("Non-Equity Scheme - Skipping")
        return {}

    eligible_transactions_filter = get_transaction_older_than_filter(num_days)

    # Get transactions for active units
    active_purchase_transactions = get_purchase_transactions_for_active_units(scheme.transactions)

    # Get eligible purchase transactions
    eligible_purchase_transactions = get_filtered_transactions(
        eligible_transactions_filter, active_purchase_transactions
    )

    total_eligible_units = sum(
        [transaction.units for transaction in eligible_purchase_transactions]
    )
    approx_amount = total_eligible_units * scheme.nav
    approx_ltcg = Decimal("0.0")

    eligible_purchase_transactions = [
        transaction.dict() for transaction in eligible_purchase_transactions
    ]

    if total_eligible_units > 0:
        for transaction in eligible_purchase_transactions:
            profit_loss = (scheme.nav - transaction["nav"]) * transaction["units"]
            approx_ltcg += profit_loss
            transaction["P&L"] = profit_loss

    return {
        "scheme": scheme.name,
        "units": total_eligible_units,
        "nav": scheme.nav,
        "amount": approx_amount,
        "ltcg": approx_ltcg,
        "transactions": eligible_purchase_transactions,
    }


def get_valuation_summary_for_schemes(schemes: List[Scheme]):
    valuation = sum([scheme.valuation for scheme in schemes])
    schemes_details = [
        {
            "name": scheme.name,
            "valuation": scheme.valuation,
            "type": scheme.type,
            "subtype": scheme.subtype,
        }
        for scheme in schemes
    ]

    return {
        "valuation": valuation,
        "schemes": schemes_details,
    }



from .models import Scheme
from .portfolio import Portfolio
from .scheme_factory import create_scheme_from_casparser

__version__ = "1.0.0"
__author__ = "Kaushik S Kalmady / kaushiksk"
__license__ = "MIT"



import datetime
from decimal import Decimal
from typing import List, Optional, Union
from casparser.types import TransactionDataType
from pydantic import BaseModel
from pydantic.annotated_types import create_model_from_typeddict

TransactionBase = create_model_from_typeddict(TransactionDataType)


class Transaction(TransactionBase):
    date: datetime.datetime
    days: int
    balance: Union[Decimal, None]


class Scheme(BaseModel):
    name: str
    amfi: str
    units: Decimal
    nav: Decimal
    type: Union[str, None]
    subtype: Union[str, None]
    valuation: Optional[Decimal]
    transactions: Optional[List[Transaction]]

    def __init__(self, **data):
        super().__init__(**data)
        self.valuation = self.units * self.nav



from typing import Callable

from .constants import DEBT, EQUITY
from .models import Scheme


SchemeFilterType = Callable[[Scheme], bool]


def get_scheme_type_filter(scheme_type: str) -> SchemeFilterType:
    def scheme_type_filter(scheme: Scheme):
        return scheme.type == scheme_type

    return scheme_type_filter


# Filters
debt_scheme_filter = get_scheme_type_filter(DEBT)
equity_scheme_filter = get_scheme_type_filter(EQUITY)



import requests

SCHEME_API_URL = "https://api.mfapi.in/mf/{code}"
session = requests.Session()


def get_scheme_details(amfi_id):
    """
    gets the scheme info for a given scheme code
    :param code: scheme code
    :return: dict or None
    :raises: HTTPError, URLError
    """
    code = str(amfi_id)
    url = SCHEME_API_URL.format(code=code)
    response = session.get(url).json()

    scheme_info = response["meta"]
    if scheme_info:
        scheme_info["nav"] = response["data"][0]["nav"]
        return scheme_info

    return None


if __name__ == "__main__":
    print(get_scheme_details(119551))
    print(get_scheme_details(111))



import casparser
from typing import List

from .scheme_utils import (
    analyze_ltcg_tax_harvesting,
    get_valuation_summary_for_schemes,
    get_filtered_schemes,
)
from .scheme_filters import debt_scheme_filter, equity_scheme_filter
from .scheme_factory import initialize_and_get_schemes
from .models import Scheme
from .utils import logger


class Portfolio:
    def __init__(self, cas_file, cas_password) -> None:
        logger.info("Parsing CAS File")
        data = casparser.read_cas_pdf(cas_file, cas_password)

        self.investor_info = data["investor_info"]

        logger.info("Loading schemes in portfolio")
        self.schemes: List[Scheme] = initialize_and_get_schemes(data)

    def ltcg_tax_harvesting_summary(self):
        equity_schemes = get_filtered_schemes(equity_scheme_filter, self.schemes)
        tax_harvesting_opportunities = [
            analyze_ltcg_tax_harvesting(scheme) for scheme in equity_schemes
        ]
        tax_harvesting_opportunities = list(
            filter(lambda x: x["ltcg"] != 0, tax_harvesting_opportunities)
        )
        total_ltcg = sum([x["ltcg"] for x in tax_harvesting_opportunities])
        total_amount = sum([x["amount"] for x in tax_harvesting_opportunities])

        return {
            "total_ltcg": total_ltcg,
            "total_amount": total_amount,
            "schemes": tax_harvesting_opportunities,
        }

    def get_valuation_summary(self):
        debt_schemes = get_filtered_schemes(debt_scheme_filter, self.schemes)
        equity_schemes = get_filtered_schemes(equity_scheme_filter, self.schemes)

        debt_valuation = get_valuation_summary_for_schemes(debt_schemes)
        equity_valuation = get_valuation_summary_for_schemes(equity_schemes)

        return {
            "valuation": debt_valuation["valuation"] + equity_valuation["valuation"],
            "debt_valuation": debt_valuation,
            "equity_valuation": equity_valuation,
        }

    def to_dict(self):
        return {
            "user_info": self.investor_info,
            "schemes": [scheme.dict() for scheme in self.schemes],
            "valuation": sum([scheme.valuation for scheme in self.schemes]),
        }



from copy import deepcopy
from decimal import Decimal
from typing import List
from .models import Transaction
from .transaction_filters import (
    TransactionFilterType,
    redemption_transaction_filter,
    purchase_transaction_filter,
)


def get_filtered_transactions(
    filter_to_apply: TransactionFilterType, transactions: List[Transaction]
):
    """Apply filter on transactions and return the filtered list"""
    return list(filter(filter_to_apply, transactions))


def get_purchase_transactions_for_active_units(transactions: List[Transaction]):
    """Returns all purchase transactions for units that haven't been redeemed yet

    Args:
        transactions (Transaction): list of transactions to check in

    Returns:
        List[Transaction]: All purchase transactions for units that haven't been redeemed
    """
    redeemed_transactions = get_filtered_transactions(redemption_transaction_filter, transactions)

    if len(redeemed_transactions) > 0:
        total_redeemed_units = abs(
            sum([transaction.units for transaction in redeemed_transactions])
        )
    else:
        total_redeemed_units = Decimal("0.0")

    # Perform deepcopy to avoid modifying the source reference
    purchase_transactions: List[Transaction] = deepcopy(
        get_filtered_transactions(purchase_transaction_filter, transactions)
    )

    if total_redeemed_units == 0:
        return purchase_transactions

    computed_units = Decimal("0.0")

    # We keep deleting transactions from top until all redemptions are satisfied
    while len(purchase_transactions) > 0:
        computed_units += purchase_transactions[0].units

        if computed_units <= total_redeemed_units:
            # All these units were extinguished during redemption
            purchase_transactions.pop(0)
        else:
            # Part of the units were extinguished
            purchase_transactions[0].units = computed_units - total_redeemed_units
            purchase_transactions[0].amount = (
                purchase_transactions[0].units * purchase_transactions[0].nav
            )
            break  # No more units were extinguished

    return purchase_transactions



import locale
import logging

import click
import cutie

from .clihelper import ltcg_tax_harvesting_summary, valuation_summary
from .portfolio import Portfolio
from .utils import logger

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(name="portfolio-cli", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-f",
    "--caspdf",
    type=click.Path(exists=True),
    metavar="CAS_PDF_FILE",
    prompt="CAS PDF File Path",
)
def main(caspdf):
    logger.setLevel(logging.INFO)
    locale.setlocale(locale.LC_MONETARY, "en_IN")

    password = cutie.secure_input("Please enter the pdf password:")

    portfolio = Portfolio(
        caspdf,
        password,
    )

    # Options
    LTCG_TAX_HARVEST = "LTGC Tax Harvesting"
    VALUATION = "Portfolio Valuation"
    EXIT = "Exit"

    options = [
        LTCG_TAX_HARVEST,
        VALUATION,
        EXIT,
    ]

    while True:
        option = options[cutie.select(options, selected_index=0)]

        if option == LTCG_TAX_HARVEST:
            ltcg_tax_harvesting_summary(portfolio)
        elif option == VALUATION:
            valuation_summary(portfolio)
        elif option == EXIT:
            break


if __name__ == "__main__":
    main(prog_name="portfolio-cli")



import logging

logging.basicConfig()
logger = logging.getLogger("pyportfolio")
logger.setLevel(logging.INFO)


# Filter Utils
def and_filter(filters):
    return lambda x: all(f(x) for f in filters)


def or_filter(filters):
    return lambda x: any(f(x) for f in filters)


def filter_dict(mydict, keys):
    return {key: mydict[key] for key in keys}



import datetime
from copy import deepcopy
from decimal import Decimal
from typing import List
from pydantic.error_wrappers import ValidationError
from casparser.types import CASParserDataType, SchemeType as CASParserSchemeType

from .models import Scheme
from .constants import EQUITY
from .mfhelper import get_scheme_details
from .utils import logger


def create_scheme_from_casparser(scheme_dict: CASParserSchemeType) -> Scheme:
    """Create an object of Scheme model from casparser data

    Args:
        scheme_dict (CASParserSchemeType): scheme dict from casparser

    Returns:
        Scheme: instance of Scheme model

    Raises:
        ValidationError is dict parsing into Scheme model fails
    """
    scheme_dict = __update_scheme_details(scheme_dict)
    try:
        scheme_model = Scheme(**scheme_dict)
        return scheme_model
    except ValidationError:
        print(
            "Dict validation failed. Make sure you are using casparser.read_cas_pdf with output = 'dict'"
        )
        raise


def __update_scheme_details(scheme_dict: CASParserSchemeType):
    """Update scheme dict from casparser with additional metadata

    Args:
        scheme_dict (CASParserSchemeType): scheme dict from casparser

    Returns:
        dict: copy of input dict with updated keys and values to conform to Scheme model
    """
    scheme_dict = deepcopy(scheme_dict)
    __update_transaction_details(scheme_dict)

    scheme_dict["units"] = scheme_dict["close_calculated"]
    scheme_dict["name"] = scheme_dict["scheme"]
    scheme_dict.pop("valuation")

    scheme_type, scheme_subtype, nav = None, None, None

    scheme_details = get_scheme_details(scheme_dict["amfi"])
    if scheme_details:
        nav = Decimal(scheme_details.get("nav"))
        scheme_type, scheme_subtype = scheme_details.get("scheme_category").split(" - ")
        if scheme_type == "Other Scheme":
            scheme_type = EQUITY

    scheme_dict["type"] = scheme_type
    scheme_dict["subtype"] = scheme_subtype
    scheme_dict["nav"] = nav

    return scheme_dict


def __update_transaction_details(scheme_dict: CASParserSchemeType):
    """Update every transaction with additional info
        - Adds a 'days' value denoting how old the transaction is
        - Transforms 'date' into datetime object to conform with Scheme model. This is to help export this data to Mongo which only supports datetime.

    Args:
        scheme_dict : opy of input dict with updated keys and values to conform to Scheme model
    """
    for transaction in scheme_dict["transactions"]:
        t_date = transaction["date"]
        transaction["days"] = (datetime.date.today() - t_date).days
        transaction["date"] = datetime.datetime(t_date.year, t_date.month, t_date.day)


def initialize_and_get_schemes(data_dict: CASParserDataType) -> List[Scheme]:
    """Process schemes in dict provided by casparser into instances of Scheme model

    Args:
        data_dict (CASParserDataType): data from casparser.read_cas_pdf

    Returns:
        List[Scheme]: List of instances of Scheme model corresponding to all schemes in provided data
    """
    schemes: List[Scheme] = []
    for folio in data_dict["folios"]:
        for scheme_dict in folio["schemes"]:
            scheme = create_scheme_from_casparser(scheme_dict)
            schemes.append(scheme)
            logger.debug(f"Done loading {scheme.name}")

    return schemes



from typing import Callable
from casparser.enums import TransactionType

from .models import Transaction
from .utils import or_filter

TransactionFilterType = Callable[[Transaction], bool]


def get_transaction_type_filter(transaction_type: TransactionType) -> TransactionFilterType:
    def transaction_type_filter(transaction: Transaction):
        return transaction.type == transaction_type

    return transaction_type_filter


def get_transaction_older_than_filter(days: int) -> TransactionFilterType:
    def transaction_older_than_filter(transaction: Transaction):
        return transaction.days > days

    return transaction_older_than_filter


# Filters
purchase_transaction_filter = or_filter(
    [
        get_transaction_type_filter(TransactionType.PURCHASE),
        get_transaction_type_filter(TransactionType.PURCHASE_SIP),
    ]
)
redemption_transaction_filter = get_transaction_type_filter(TransactionType.REDEMPTION)
stampduty_transaction_filter = get_transaction_type_filter(TransactionType.MISC)



from decimal import Decimal
import locale
from collections import defaultdict

import click
import cutie
import tabulate

from .portfolio import Portfolio


def ltcg_tax_harvesting_summary(portfolio: Portfolio):
    summary = portfolio.ltcg_tax_harvesting_summary()

    print(
        click.style("\nTotal LTCG Tax that can be harvested this FY: ", bold=True)
        + click.style(locale.currency(summary["total_ltcg"], grouping=True), bold=True, fg="green")
    )
    print(
        click.style("Total value of units to be sold and bought: ", bold=True)
        + click.style(
            locale.currency(summary["total_amount"], grouping=True),
            bold=True,
            fg="green",
        )
    )
    print(
        click.style(
            """Note: Buy and sell the following units on the same day to avoid losses due to NAV
            fluctuations""",
            bold=True,
            fg="yellow",
        )
    )

    schemes_breakup = summary["schemes"]
    schemes_breakup = [
        {k: v for k, v in scheme.items() if k != "transactions"} for scheme in schemes_breakup
    ]
    rows = [x.values() for x in schemes_breakup]
    header = schemes_breakup[0].keys()
    print(tabulate.tabulate(rows, header, tablefmt="fancy_grid"))

    if cutie.prompt_yes_or_no("Show scheme wise break-up of eligible transactions?"):
        schemes_breakup = summary["schemes"]
        for scheme in schemes_breakup:
            eligible_purchase_transactions = scheme["transactions"]
            print(click.style(scheme["scheme"], bold=True))
            print(
                "Total units available for LTCG tax harvesting: "
                + click.style(str(scheme["units"]), bold=True)
            )
            print(
                f"Approx value at NAV {scheme['nav']}: "
                + click.style(
                    locale.currency(scheme["amount"], grouping=True),
                    bold=True,
                    fg="green",
                )
            )
            print(
                "Approx LTCG: "
                + click.style(
                    locale.currency(scheme["ltcg"], grouping=True),
                    bold=True,
                    fg="green",
                )
            )

            print("\nBreak-down: ")
            rows = [x.values() for x in eligible_purchase_transactions]
            header = eligible_purchase_transactions[0].keys()
            print(tabulate.tabulate(rows, header, tablefmt="grid"))
            print("\n")


def valuation_summary(portfolio: Portfolio):
    summary = portfolio.get_valuation_summary()
    debt_summary = summary["debt_valuation"]
    equity_summary = summary["equity_valuation"]

    total_valuation = summary["valuation"]
    debt_percentage = (debt_summary["valuation"] * 100) / total_valuation
    equity_percentage = (equity_summary["valuation"] * 100) / total_valuation

    print(
        click.style("\nValuation: ", bold=True)
        + click.style(locale.currency(total_valuation, grouping=True), bold=True, fg="green")
    )
    print(click.style("Debt  : " + " {0:.2f}%".format(debt_percentage), bold=True))
    print(click.style("Equity: " + " {0:.2f}%".format(equity_percentage), bold=True))

    if cutie.prompt_yes_or_no("Show subtype breakup for debt and equity?"):
        print()

        def print_subtypes_summary(summary, title):
            subtypes = defaultdict(Decimal)
            subtype_counts = defaultdict(int)
            for scheme in summary["schemes"]:
                subtypes[scheme["subtype"]] += scheme["valuation"]
                subtype_counts[scheme["subtype"]] += 1

            subtypes = sorted(subtypes.items(), key=lambda x: x[1], reverse=True)

            valuation = summary["valuation"]
            valuation_percentage = (valuation * 100) / total_valuation

            print(
                click.style(
                    "{0:<6} - ".format(title)
                    + click.style(locale.currency(valuation, grouping=True), fg="green")
                    + " ({0:.2f}%)".format(valuation_percentage),
                    bold=True,
                )
            )
            for subtype, subtype_valuation in subtypes:
                print(
                    click.style(
                        " {0:<25} - {1} ({2:.2f}%)".format(
                            subtype + " ({})".format(subtype_counts[subtype]),
                            click.style(
                                locale.currency(subtype_valuation, grouping=True),
                                fg="green",
                            ),
                            (subtype_valuation * 100) / valuation,
                        ),
                        bold=True,
                    )
                )

            print()

        print_subtypes_summary(debt_summary, "Debt")
        print_subtypes_summary(equity_summary, "Equity")
